"""
Import python modules with bpftrace supervision.
Copyright (c) 2022 Avi Lumelsky
"""

import importlib
import os
from sys import executable as PYTHON_EXECUTABLE
import stat
import time
from pathlib import Path
from typing import List

from secimport.backends.common.instrumentation_backend import InstrumentationBackend
from secimport.backends.common.utils import (
    BASE_DIR_NAME,
    SECIMPORT_ROOT,
    render_syscalls_filter,
)

TEMPLATES_DIR_NAME = SECIMPORT_ROOT / "backends" / "bpftrace_backend"


def create_bpftrace_script_for_module(
    module_name: str,
    allow_shells: bool,
    allow_networking: bool,
    log_python_calls: bool,
    log_syscalls: bool,
    log_network: bool,
    log_file_system: bool,
    destructive: bool,
    syscalls_allowlist: List[str] = None,
    syscalls_blocklist: List[str] = None,
    templates_dir: Path = TEMPLATES_DIR_NAME,
) -> Path:
    """
    # Template components available at the moment:
    # ###DESTRUCTIVE###
    # ###FUNCTION_ENTRY###
    # ###FUNCTION_EXIT###
    # ###SYSCALL_ENTRY###
    # ###MODULE_NAME###
    """

    module = importlib.machinery.PathFinder().find_spec(module_name)
    if module is None:
        raise ModuleNotFoundError(module_name)
    module_traced_name = module.origin  # e.g this.py

    assert not (
        syscalls_allowlist is not None and syscalls_blocklist is not None
    ), "Please specify either syscalls_allowlist OR syscalls_blocklist."

    # If we have an allowlist
    if syscalls_allowlist is not None:
        script_template = render_bpftrace_probe_for_module(
            module_name=module_name,
            destructive=destructive,
            templates_dir=templates_dir,
        )
        syscalls_filter = render_syscalls_filter(
            syscalls_list=syscalls_allowlist,
            allow=True,
            instrumentation_backend=InstrumentationBackend.EBPF,
            module_name=module_traced_name,
        )
        script_template = script_template.replace(
            "###SYSCALL_FILTER###", syscalls_filter
        )

    elif syscalls_blocklist is not None:
        # If we have a blocklist
        script_template = render_bpftrace_probe_for_module(
            module_name=module_name,
            destructive=destructive,
            templates_dir=templates_dir,
        )
        syscalls_filter = render_syscalls_filter(
            syscalls_list=syscalls_allowlist,
            allow=True,
            instrumentation_backend=InstrumentationBackend.EBPF,
            module_name=module_traced_name,
        )
        script_template = script_template.replace(
            "###SYSCALL_FILTER###", syscalls_filter
        )

    else:
        script_template = render_bpftrace_template(
            module_traced_name=module_traced_name,
            allow_shells=allow_shells,
            allow_networking=allow_networking,
            log_python_calls=log_python_calls,
            log_syscalls=log_syscalls,
            log_network=log_network,
            log_file_system=log_file_system,
            destructive=destructive,
            templates_dir=templates_dir,
        )

    # Creating a dscript file with the modified template
    if not os.path.exists(BASE_DIR_NAME):
        os.mkdir(BASE_DIR_NAME)

    module_file_name = os.path.join(BASE_DIR_NAME, f"bpftrace_sandbox_{module_name}.bt")
    with open(module_file_name, "w") as module_file:
        module_file.write(script_template)

    # TODO: FIGURE OUT A WAY TO COMPILE WITHOUT EXECUTING - MSKING SURE THE GENERATION WORKED SYNTAX-WISE.
    return module_file_name


def run_bpftrace_script_for_module(
    module_name: str,
    allow_shells: bool,
    allow_networking: bool,
    log_python_calls: bool,
    log_syscalls: bool,
    log_network: bool,
    log_file_system: bool,
    destructive: bool,
    syscalls_allowlist: List[str],
    syscalls_blocklist: List[str],
    use_sudo: bool = False,
    templates_dir: Path = TEMPLATES_DIR_NAME,
):
    module_file_path = create_bpftrace_script_for_module(
        module_name=module_name,
        allow_shells=allow_shells,
        allow_networking=allow_networking,
        log_python_calls=log_python_calls,
        log_syscalls=log_syscalls,
        log_network=log_network,
        log_file_system=log_file_system,
        destructive=destructive,
        syscalls_allowlist=syscalls_allowlist,
        syscalls_blocklist=syscalls_blocklist,
        templates_dir=templates_dir,
    )
    output_file = BASE_DIR_NAME / f"bpftrace_sandbox_{module_name}.log"
    current_pid = os.getpid()
    bpftrace_command = f'{"sudo " if use_sudo else ""} {module_file_path} --unsafe -q -p {current_pid} -o {output_file} &2>/dev/null'
    st = os.stat(module_file_path)
    os.chmod(module_file_path, st.st_mode | stat.S_IEXEC)
    print("(sandbox command): ", bpftrace_command)
    print()
    print("Waiting for bpftrace.... ")
    os.system(bpftrace_command)
    time.sleep(10)  # TODO: change to fd creation (event)
    return True


def render_bpftrace_template(
    module_traced_name: str,
    allow_shells: bool,
    allow_networking: bool,
    log_python_calls: bool,
    log_syscalls: bool,
    log_network: bool,
    log_file_system: bool,
    destructive: bool,
    templates_dir: Path = TEMPLATES_DIR_NAME,
    default_template_filename: str = "default.template.bt",
    interpreter_path: str = PYTHON_EXECUTABLE,
):
    script_template = open(
        templates_dir / default_template_filename,
        "r",
    ).read()

    # Updating the right binary to be probed by bpftrace
    script_template = script_template.replace(
        "###INTERPRETER_PATH###", interpreter_path
    )

    # PYTHON instrumentations
    code_syscall_entry = ""
    if log_python_calls is True:
        code_function_entry = open(
            templates_dir / "actions/log_python_module_entry.bt",
            "r",
        ).read()
        code_function_exit = open(
            templates_dir / "actions/log_python_module_exit.bt",
            "r",
        ).read()
        script_template = script_template.replace(
            "###FUNCTION_ENTRY###", code_function_entry
        )
        script_template = script_template.replace(
            "###FUNCTION_EXIT###", code_function_exit
        )
    else:
        script_template = script_template.replace("###FUNCTION_ENTRY###", "")
        script_template = script_template.replace("###FUNCTION_EXIT###", "")

    # SYSCALLS instrumentations
    if log_syscalls is True:
        _code = open(
            templates_dir / "actions/log_syscall.bt",
            "r",
        ).read()
        code_syscall_entry += f"{_code};\n"

    if log_file_system is True:
        filter_fs_code = open(
            templates_dir / "filters/file_system.bt",
            "r",
        ).read()
        code_syscall_entry += f"{filter_fs_code}\n{{"
        action_log_file_system = open(
            templates_dir / "actions/log_file_system.bt",
            "r",
        ).read()
        code_syscall_entry += f"{action_log_file_system}}}\n"

    if allow_networking is False or log_network is True:
        filter_networking_code = open(
            templates_dir / "filters/networking.bt",
            "r",
        ).read()
        code_syscall_entry += f"{filter_networking_code}{{\n"

        if log_network is True:
            action_log_network = open(
                templates_dir / "actions/log_network.bt",
                "r",
            ).read()
            code_syscall_entry += f"{action_log_network}\n"

        if allow_networking is False:
            action_kill = open(
                templates_dir / "actions/kill_process.bt",
                "r",
            ).read()
            code_syscall_entry += f"{action_kill}\n"
        code_syscall_entry += "}\n"

    if allow_shells is False:
        filter_processes_code = open(
            templates_dir / "filters/processes.bt",
            "r",
        ).read()
        code_syscall_entry += f"{filter_processes_code}{{\n"
        action_kill_on_processing = open(
            templates_dir / "actions/kill_on_processing.bt",
            "r",
        ).read()
        code_syscall_entry += f"{action_kill_on_processing}}}\n"

    script_template = script_template.replace("###SYSCALL_ENTRY###", code_syscall_entry)
    script_template = script_template.replace("###MODULE_NAME###", module_traced_name)

    # Updating the destructive behavior
    script_template = script_template.replace(
        "###DESTRUCTIVE###", "1" if destructive else "0"
    )
    return script_template


def render_bpftrace_probe_for_module(
    module_name: str,
    destructive: bool,
    templates_dir: Path = TEMPLATES_DIR_NAME,
    interpreter_path: str = PYTHON_EXECUTABLE,
) -> str:
    # Loading the probe allowlist template
    probe_template = open(
        templates_dir / "probes/module_syscalls_allowlist_template.bt",
        "r",
    ).read()

    # Adding a probe filter for the specified python module
    supervision_filter = open(
        templates_dir / "filters/is_current_module_under_supervision.bt",
        "r",
    ).read()

    # Adding the action according to the 'destructive' flag:  kill if destructive, log otherwise
    supervision_action = ""
    if destructive is True:
        action_kill_process = open(
            templates_dir / "actions/kill_process.bt",
            "r",
        ).read()
        supervision_action = f"{{{action_kill_process}}}\n"
    else:
        action_log_syscall = open(
            templates_dir / "actions/log_syscall.bt",
            "r",
        ).read()
        supervision_action = f"{{{action_log_syscall}}}\n"

    # Updating the template with the filters, their actions, and module name
    probe_template = probe_template.replace(
        "###SUPERVISED_MODULES_FILTER###", supervision_filter
    )
    probe_template = probe_template.replace(
        "###SUPERVISED_MODULES_ACTION###", supervision_action
    )
    probe_template = probe_template.replace("###MODULE_NAME###", module_name)

    # Updating the destructive behavior
    probe_template = probe_template.replace(
        "###DESTRUCTIVE###", "1" if destructive else "0"
    )

    probe_template = probe_template.replace("###INTERPRETER_PATH###", interpreter_path)
    return probe_template



"""
Import python modules with syscalls supervision.

Copyright (c) 2022 Avi Lumelsky
"""

from typing import List
from secimport.backends.common.instrumentation_backend import InstrumentationBackend
from secimport.backends.common.utils import DEFAULT_BACKEND
from secimport.backends.bpftrace_backend.bpftrace_backend import (
    run_bpftrace_script_for_module,
)
from secimport.backends.dtrace_backend.dtrace_backend import (
    run_dtrace_script_for_module,
)


def secure_import(
    module_name: str,
    allow_shells: bool = False,
    allow_networking: bool = False,
    use_sudo: bool = False,
    log_python_calls: bool = False,  # When True, the log file might reach GB in seconds.
    log_syscalls: bool = False,
    log_network: bool = False,
    log_file_system: bool = False,
    destructive: bool = True,
    syscalls_allowlist: List[str] = None,
    syscalls_blocklist: List[str] = None,
    backend=DEFAULT_BACKEND,
):
    """Import a python module in confined settings.

    Args:
        module_name (str): The name of the module to import. The same way you would have used an 'import' statement.
        allow_shells (bool, optional): Whether to allow forking/spwning shells and execute commands.
        allow_networking (bool, optional): Whether to allow socket syscalls.
        use_sudo (bool, optional): Whether to run dtrace with sudo. Can be turned off if sudo is not installed.
        log_python_calls (bool, optional): Whether to log the call tree of the python stack - function entry and exit events.
        log_network (bool, optional): Whether to log network successful connections - e.g socket syscalls.
        log_file_system (bool, optional): Whether to log filesystem calls - e.g read, open, write, etc.
        destructive (bool, optional): Whether to kill the process with -9 sigkill upon violation of any of the configurations above.
        syscalls_allowlist (list, optional): A list of allowed syscalls for the module. Any other syscall will result in destructive behavior (see description above).
        syscalls_blocklist (list, optional): A list of syscalls to block for the module. Any of the specified syscalls under this module will result in destructive behavior (see description above).
    Returns:
        _type_: A Python Module. The module is supervised by a dtrace process with destructive capabilities unless the 'destructive' argument is set to False.
    """

    if backend == InstrumentationBackend.EBPF:
        assert run_bpftrace_script_for_module(
            module_name=module_name,
            allow_shells=allow_shells,
            allow_networking=allow_networking,
            use_sudo=use_sudo,
            log_python_calls=log_python_calls,
            log_syscalls=log_syscalls,
            log_network=log_network,
            log_file_system=log_file_system,
            destructive=destructive,
            syscalls_allowlist=syscalls_allowlist,
            syscalls_blocklist=syscalls_blocklist,
        )
    elif backend == InstrumentationBackend.DTRACE:
        assert run_dtrace_script_for_module(
            module_name=module_name,
            allow_shells=allow_shells,
            allow_networking=allow_networking,
            use_sudo=use_sudo,
            log_python_calls=log_python_calls,
            log_syscalls=log_syscalls,
            log_network=log_network,
            log_file_system=log_file_system,
            destructive=destructive,
            syscalls_allowlist=syscalls_allowlist,
            syscalls_blocklist=syscalls_blocklist,
        )
    else:
        raise NotImplementedError(f"backend '{backend}' is not implemented.")

    _module = __import__(module_name)
    return _module



"""
An abstract backend for language runtime instrumentation.
Copyright (c) 2022 Avi Lumelsky
"""

from abc import ABC
from sys import executable as PYTHON_EXECUTABLE
from pathlib import Path
from typing import List

from secimport.backends.common.instrumentation_backend import InstrumentationBackend
from secimport.backends.common.utils import TEMPLATES_DIR_NAME


class SandboxInstrumentationBackend(ABC):
    @property
    def name(self):
        # return "bpftrace"
        raise NotImplementedError()

    @property
    def backend(self):
        # return InstrumentationBackend.EBPF
        raise NotImplementedError()

    @staticmethod
    def run_script_for_module(
        module_name: str,
        allow_shells: bool,
        allow_networking: bool,
        log_python_calls: bool,
        log_syscalls: bool,
        log_network: bool,
        log_file_system: bool,
        destructive: bool,
        syscalls_allowlist: List[str],
        syscalls_blocklist: List[str],
        use_sudo: bool = False,
        templates_dir: Path = TEMPLATES_DIR_NAME,
    ):
        raise NotImplementedError()

    @staticmethod
    def create_script_for_module(
        module_name: str,
        allow_shells: bool,
        allow_networking: bool,
        log_python_calls: bool,
        log_syscalls: bool,
        log_network: bool,
        log_file_system: bool,
        destructive: bool,
        syscalls_allowlist: List[str] = None,
        syscalls_blocklist: List[str] = None,
        templates_dir: Path = TEMPLATES_DIR_NAME,
    ) -> Path:
        raise NotImplementedError()

    @staticmethod
    def render_bpftrace_template(
        module_traced_name: str,
        allow_shells: bool,
        allow_networking: bool,
        log_python_calls: bool,
        log_syscalls: bool,
        log_network: bool,
        log_file_system: bool,
        destructive: bool,
        templates_dir: Path = TEMPLATES_DIR_NAME,
        default_template_filename: str = "default.template.example",
        interpreter_path: str = PYTHON_EXECUTABLE,
    ):
        raise NotImplementedError()

    @staticmethod
    def render_allowlist_template(
        module_name: str,
        destructive: bool,
        syscalls_allowlist: List[str],
        templates_dir: Path = TEMPLATES_DIR_NAME,
    ):
        raise NotImplementedError()

    @staticmethod
    def render_blocklist_template(
        module_name: str,
        destructive: bool,
        syscalls_blocklist: List[str],
        templates_dir: Path = TEMPLATES_DIR_NAME,
    ):
        raise NotImplementedError()

    @staticmethod
    def render_instrumentation_for_module(
        module_name: str,
        destructive: bool,
        syscalls_list: List[str],
        syscalls_allow: bool,
        templates_dir: Path = TEMPLATES_DIR_NAME,
        interpreter_path: str = PYTHON_EXECUTABLE,
    ) -> str:
        raise NotImplementedError()



from collections import defaultdict
import json
import os
from pathlib import Path
import stat
import fire
import sys
import yaml
import secimport
from secimport.backends.common.utils import SECIMPORT_ROOT
from secimport.sandbox_helper import secure_import
import subprocess


class SecImportCLI:
    """Secure Import CLI"""

    @staticmethod
    def trace(
        python_interpreter: str = sys.executable,
        trace_log_file: str = "trace.log",
    ):
        print("To start a trace, please run the following command:")
        print(
            f"{SECIMPORT_ROOT}/profiles/trace.bt -c '{python_interpreter}' -o {trace_log_file}"
        )

    @staticmethod
    def trace_pid(
        pid: int,
        trace_log_file: str = "trace.log",
    ):
        """Traces a running process by pid. It might require sudo privilleges on some hosts.

        Args:
            pid (int): process pid to trace
            trace_log_file (str): The output file to write the trace into.
        """
        assert isinstance(pid, int), f"pid must be an int, got: {pid}"
        cmd = [
            SECIMPORT_ROOT + "/profiles/trace.bt",
            "-p",
            "{pid}",
            "-o",
            f"{trace_log_file}",
        ]
        print("Running trace command:")
        print("".join(cmd))
        process = subprocess.run(cmd)
        print(process)
        input("Press CTRL+D or CTRL+C to exit the trace...")

    @staticmethod
    def compile_sandbox(trace_file: str = "trace.log") -> str:
        """Converts a trace log into sandbox. It uses `secimport create_profile_from_trace ...` and `secimport sandbox_from_profile...` under the hood.

        Raises:
            FileNotFoundError:

        Returns:
            _type_: str
        """
        output_yaml_file = SecImportCLI.create_profile_from_trace(trace_file=trace_file)
        bpftrace_sandbox_filename = SecImportCLI.create_sandbox_from_profile(
            template_filename=output_yaml_file
        )
        return bpftrace_sandbox_filename

    @staticmethod
    def create_profile_from_trace(
        trace_file: str = "trace.log",
        output_json_file: str = "traced_modules.json",
        output_yaml_file: str = "traced_modules.yaml",
    ) -> str:
        """Creates a yaml policy file from a given trace log. default: "trace.log".

        After running this command, the yaml output file can be compiled into sandbox code using the 'sandbox_from_profile' command.
        """
        parse_cmd = f"""cat {trace_file} | grep -Po "\@modules_syscalls\[\K(.+?)]" | tr -d ']' | sort > traced_modules.log"""
        os.system(parse_cmd)
        modules_syscalls_file = Path("traced_modules.log")
        if not modules_syscalls_file.exists():
            raise FileNotFoundError(trace_file)
        with open(modules_syscalls_file, "r") as f:
            lines = f.read().splitlines()

        modules_and_syscalls = defaultdict(list)
        for line in lines:
            parsed_items = line.split(",")
            if parsed_items == 1:
                # General python import, interpreter dependency
                modules_and_syscalls["general_requirements"].append(_syscall)
            elif len(parsed_items) == 2:
                # Module dependency
                _module, _syscall = parsed_items
                if _module == "":
                    _module = "general_requirements"
                modules_and_syscalls[_module].append(_syscall)
            else:
                raise Exception("there are more than 2 items in one of the lines.")

        # JSON
        modules_json = json.dumps(modules_and_syscalls, indent=4)
        with open(output_json_file, "w") as f:
            f.write(modules_json)
        print("The JSON trace is ready at ", output_json_file)

        # Building the YAML template
        yaml_dict = {
            name: {
                "destructive": True,
                "syscall_allowlist": _syscalls,
            }
            for name, _syscalls in modules_and_syscalls.items()
        }
        yaml_dict = {"modules": yaml_dict}
        modules_yaml = yaml.safe_load(json.dumps(yaml_dict))
        modules_yaml = yaml.dump(modules_yaml)
        with open(output_yaml_file, "w") as f:
            f.write(modules_yaml)
        print("The YAML template is ready at ", output_yaml_file)
        return output_yaml_file

    @staticmethod
    def create_sandbox_from_profile(
        template_filename: str = "traced_modules.yaml",
    ) -> str:
        """Generates a tailor-made sandbox code for a given yaml profile.

        2 scripts will be created in the same directory: one for the dscript backend and one for the bpftrace backend.
        Returns: The file name of the bpftrace sandbox (preferred).
            _type_: str
        """
        print("Compiling template", template_filename)
        template_filename = Path(template_filename)
        dtrace_sandbox_filename = Path(template_filename).with_suffix(".d")
        dtrace_rendered_profile = secimport.build_module_sandbox_from_yaml_template(
            template_path=template_filename,
            backend=secimport.InstrumentationBackend.DTRACE,
        )
        with open(dtrace_sandbox_filename, "w") as f:
            f.write(dtrace_rendered_profile)
        st = os.stat(dtrace_sandbox_filename)
        os.chmod(dtrace_sandbox_filename, st.st_mode | stat.S_IEXEC)
        bpftrace_sandbox_filename = Path(template_filename).with_suffix(".bt")
        bpftrace_rendered_profile = secimport.build_module_sandbox_from_yaml_template(
            template_path=template_filename,
            backend=secimport.InstrumentationBackend.EBPF,
        )
        with open(bpftrace_sandbox_filename, "w") as f:
            f.write(bpftrace_rendered_profile)
        st = os.stat(bpftrace_sandbox_filename)
        os.chmod(bpftrace_sandbox_filename, st.st_mode | stat.S_IEXEC)

        print("The dtrace sandbox script is ready at ", dtrace_sandbox_filename)
        print("The bpftrace sandbox script is ready at ", bpftrace_sandbox_filename)
        print("Done.")
        return bpftrace_sandbox_filename

    @staticmethod
    def secure_import():
        """Run a secure import for a python module and then exit.
        A log will be automatically generated at /tmp/.secimport"""

        return fire.Fire(secure_import)


def main():
    fire.Fire(SecImportCLI)


if __name__ == "__main__":
    main()



import os
from sys import platform
from pathlib import Path
from typing import List
import importlib
from sys import executable as PYTHON_EXECUTABLE

from secimport.backends.common.instrumentation_backend import InstrumentationBackend

BASE_DIR_NAME = Path("/tmp/.secimport")
SECIMPORT_ROOT = Path(
    os.path.realpath(
        os.path.split(__file__)[:-1][0] + os.sep + os.pardir + os.sep + os.pardir
    )
)

PROFILES_DIR_NAME = SECIMPORT_ROOT / "profiles"
BPFTRACE_TEMPLATES_DIR_NAME = SECIMPORT_ROOT / "backends" / "bpftrace_backend"
DTRACE_TEMPLATES_DIR_NAME = SECIMPORT_ROOT / "backends" / "dtrace_backend"
TEMPLATES_DIR_NAME = None
DEFAULT_BACKEND = None

if "linux" in platform.lower():
    TEMPLATES_DIR_NAME = BPFTRACE_TEMPLATES_DIR_NAME
    DEFAULT_BACKEND = InstrumentationBackend.EBPF
    # TODO: verify bpftrace is installed, if not, link to the repo documentation on how to install.
else:
    TEMPLATES_DIR_NAME = DTRACE_TEMPLATES_DIR_NAME
    DEFAULT_BACKEND = InstrumentationBackend.DTRACE


def render_syscalls_filter(
    syscalls_list: List[str],
    allow: bool,
    instrumentation_backend: InstrumentationBackend,
    module_name=None,
):
    assert isinstance(allow, bool), '"allow" must be a bool value'
    # "=="" means the syscall matches (blocklist), while "!="" means allow only the following.
    match_sign = "!=" if allow else "=="
    syscalls_filter = ""
    for i, _syscall in enumerate(syscalls_list):
        if i > 0:
            if instrumentation_backend == InstrumentationBackend.DTRACE:
                syscalls_filter += " && "
        assert isinstance(
            _syscall, str
        ), f"The provided syscall it not a syscall string name: {_syscall}"

        if instrumentation_backend == InstrumentationBackend.DTRACE:
            if module_name is not None:
                raise NotImplementedError(
                    "Specific modules syscalls are not allowed using render_syscall_filter. Please use YAML template instead."
                )
            syscalls_filter += f'probefunc {match_sign} "{_syscall}"'
        elif instrumentation_backend == InstrumentationBackend.EBPF:
            # We denote 1 as 'allow' and 0 as 'deny', 0 being the deafult or unset.
            expected_output_value = "1" if allow else "2"
            # strncmp returns 0 if it is a match
            # syscalls_filter += f'(strncmp(@sysname[args->id],  "{_syscall}", 24) {expected_output_value})'
            _module_name = (
                f'"{module_name}"'
                if module_name is not None
                else '@globals["current_module"]'
            )
            syscalls_filter += f"@syscalls_filters[{_module_name}, \"{_syscall}\"] = {expected_output_value};"
            syscalls_filter += "\n\t"
        else:
            raise NotImplementedError(
                f"backend '{instrumentation_backend}' is not supported"
            )

        filter_name = "allowlist" if allow else "blocklist"
        print(f"Adding syscall {_syscall} to {filter_name} for module {module_name}")
    return syscalls_filter


def build_module_sandbox_from_yaml_template(
    template_path: Path,
    backend: InstrumentationBackend = DEFAULT_BACKEND,
):
    """Generates sandbox code for secure imports based on a YAML configuration.

    Args:
        template_path (Path): The path to the YAML file, describing the policies.
        templates_dir (Path, optional): The directory of the templates. Defaults to TEMPLATES_DIR_NAME.

    Raises:
        ModuleNotFoundError:

    Returns:
        _type_: str
    """
    assert Path(
        template_path
    ).exists(), f"The template does not exist at {template_path}"
    import yaml

    safe_yaml = yaml.safe_load(open(template_path, "r").read())
    parsed_probes = []
    syscalls_filter = ""
    for module_name, module_config in safe_yaml.get("modules", {}).items():
        # Finding the module without loading
        module = importlib.machinery.PathFinder().find_spec(module_name)
        if module is None:
            msg = f"Warning: {module_name} is not present in the current environment. It is required in order to generate a sandbox with correct a import name."
            # raise ModuleNotFoundError(msg)

        # Tracing module entrypoint
        module_traced_name = module.origin if module is not None else module_name
        # module_traced_name = os.path.split(module_traced_name)[:-1][0]

        _destructive = module_config.get("destructive")
        assert isinstance(_destructive, bool), ValueError(
            f'The "destructive" field for module {module_name} is empty.'
        )

        _syscall_allowlist = module_config.get("syscall_allowlist")
        assert _syscall_allowlist, ValueError(
            f'The "syscall_allowlist" for module {module_name} is empty.'
        )
        for _ in _syscall_allowlist:
            assert isinstance(_, str), ValueError(
                f'The "syscall_allowlist" field for module {module_name} contains invalid string: {_}'
            )

        if backend == InstrumentationBackend.DTRACE:
            from secimport.backends.dtrace_backend.dtrace_backend import (
                render_dtrace_probe_for_module,
            )

            module_sandbox_probe = render_dtrace_probe_for_module(
                module_name=module_traced_name,
                destructive=_destructive,
                syscalls_allowlist=_syscall_allowlist,
            )
            sandbox_file_name = f"default.yaml.template.d"
            script_template = open(
                DTRACE_TEMPLATES_DIR_NAME / sandbox_file_name,
                "r",
            ).read()
        elif backend == InstrumentationBackend.EBPF:
            from secimport.backends.bpftrace_backend.bpftrace_backend import (
                render_bpftrace_probe_for_module,
            )

            module_sandbox_probe = render_bpftrace_probe_for_module(
                module_name=module_traced_name,
                destructive=_destructive,
            )
            # Adding a syscalls filter
            syscalls_filter += render_syscalls_filter(
                syscalls_list=_syscall_allowlist,
                allow=True,
                instrumentation_backend=InstrumentationBackend.EBPF,
                module_name=module_traced_name,
            )
            sandbox_file_name = f"default.yaml.template.bt"
            script_template = open(
                BPFTRACE_TEMPLATES_DIR_NAME / sandbox_file_name,
                "r",
            ).read()
        assert module_sandbox_probe, ValueError(
            f"Failed to create a probe for module {module_name}"
        )
        parsed_probes.append(module_sandbox_probe)

    if not parsed_probes:
        print(f"The profile does not contain any modules: {template_path}")
        return
    script_template = script_template.replace("###SYSCALL_FILTER###", syscalls_filter)
    script_template = script_template.replace(
        "###INTERPRETER_PATH###", PYTHON_EXECUTABLE
    )
    probes_code = ("\n" * 2).join(parsed_probes)
    script_template = script_template.replace(
        "###SUPERVISED_MODULES_PROBES###", probes_code
    )
    return script_template



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['secimport',
 'secimport.backends.bpftrace_backend',
 'secimport.backends.common',
 'secimport.backends.dtrace_backend']

package_data = \
{'': ['*'],
 'secimport': ['profiles/*'],
 'secimport.backends.bpftrace_backend': ['actions/*', 'filters/*', 'probes/*'],
 'secimport.backends.dtrace_backend': ['actions/*',
                                       'filters/*',
                                       'headers/*',
                                       'probes/*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'fire>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['secimport = secimport.cli:main']}

setup_kwargs = {
    'name': 'secimport',
    'version': '0.6.0',
    'description': 'A sandbox/supervisor for python modules.',
    'long_description': '# secimport\nA sandbox that grants privilleges per module in your code, making 3rd pary and open source and untrusted code to run in trusted environments.<br>\nIt does using backends like bpftrace (eBPF) and dtrace under the hood.<br>\n<p align="center">\n<a href="https://infosecwriteups.com/sandboxing-python-modules-in-your-code-1e590d71fc26?source=friends_link&sk=5e9a2fa4d4921af0ec94f175f7ee49f9">Medium Article</a>\n</p>\n<p align="center">\n <a href="https://github.com/avilum/secimport"><img style="max-height: 100px" src="https://user-images.githubusercontent.com/19243302/177835749-6aec7200-718e-431a-9ab5-c83c6f68565e.png" alt="secimport"></a>\n</p>\n\n# How is works?\n`secimport` uses USDT (Userland Statically Defined Tracing) probes in the runtime (Python interpreter for example). OS kernel using eBPF and dtrace instrumentation scripts.<br>\nYou can use `secimport` to:\n- Trace which syscalls are called by each module in your code, or by your entire application.\n- Restrict specific modules/packages inside your production environment like 3rd party, open source or code from untrusted source.\n- Audit the flow of your python application at user-space/os/kernel level.\n- Kill or audit upon violoation of your specified behavior. Killing is optional.\n<br><br>\n# Quick Start\nThere are several methods to create and run a sandbox:\n1. By modifying your imports\n    - Inside your code using `module = secimport.secure_import(\'module_name\', ...)`.\n      - Replacing the regular `import` statement with `secure_import`\n      - Only modules that were imported with `secure_import` will be traced.\n2. By running it as a parent process for your application\n      -  Generate a YAML policy from your code, by specifying the modules and the policy you want, for every module that you would like to restrict in any way.\n         - Convert that YAML policy to dscript/bpftrace sandbox code.\n      - Run you tailor-made sandbox\n          - Use `dtrace` or `bpftrace` to run your main python application, with your tailor-made sandbox.\n          - No need for `secure_import`, you can keep using regular `import`s and not change your code at all.\n<br><br>\n# Docker\nThe easiest way to try secimport is by using our <a href="docker/README.md">Docker for MacOS and Linux</a>. It includes python, secimport and bpftrace backend.<br>\n`dtrace` backend is not available in docker, and can be tried directly on the compatible hosts ( <a href="docs/MAC_OS_USERS.md">Mac OS</a> , Windows, Solaris, Unix, some Linux distributions).\n<br><br>\n\n# References:\n- Read about the available backends in secimport:\n  - https://www.brendangregg.com/DTrace/DTrace-cheatsheet.pdf\n    - `dtrace`\n  - https://www.brendangregg.com/blog/2018-10-08/dtrace-for-linux-2018.html\n    - `bpftrace` (dtrace 2.0) that uses LLVM and compiled our script to BCC.\n       - https://github.com/iovisor/bpftrace\n- <a href="docs/EXAMPLES.md">Examples</a>\n\n- Guides\n  - <a href="docs/TRACING_PROCESSES.md">Tracing Processes Guide</a>\n  - <a href="docs/INSTALL.md">Installation</a>\n  - <a href="docs/MAC_OS_USERS.md">Mac OS Users</a> - Disabling SIP (System Intergity Protection)\n  - <a href="docs/FAQ.md">F.A.Q</a>\n  <br><br>\n\n\n# Example Use Cases\n<a href="docs/EXAMPLES.md">EXAMPLES.md</a> contains advanced usage and many interactive session examples: YAML profies, networking, filesystem, processing blocking & more.\n\n## Simple Usage\n- <a href="examples/python_imports/">Running Sandbox Using Python Imports</a>\n- <a href="docs/CLI.md">`secimport` CLI usage</a>\n    - The easiest option to start with inside docker.\n    - `python -m secimport.cli --help`\n- See <a href="YAML_PROFILES.md">YAML Profiles Usage</a>>\n<br><br>\n### How pickle can be exploited in your 3rd party packages (and how to block it)\n```python\n# Not your code, but you load and run it frmo 3rd some party package.\n\nimport pickle\nclass Demo:\n    def __reduce__(self):\n        return (eval, ("__import__(\'os\').system(\'echo Exploited!\')",))\n \npickle.dumps(Demo())\nb"\\x80\\x04\\x95F\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x8c\\x08builtins\\x94\\x8c\\x04eval\\x94\\x93\\x94\\x8c*__import__(\'os\').system(\'echo Exploited!\')\\x94\\x85\\x94R\\x94."\n\n# Your code, at some day...\npickle.loads(b"\\x80\\x04\\x95F\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x8c\\x08builtins\\x94\\x8c\\x04eval\\x94\\x93\\x94\\x8c*__import__(\'os\').system(\'echo Exploited!\')\\x94\\x85\\x94R\\x94.")\nExploited!\n0\n```\nWith `secimport`, you can control such action to do whatever you want:\n```python\nimport secimport\npickle = secimport.secure_import("pickle")\npickle.loads(b"\\x80\\x04\\x95F\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x8c\\x08builtins\\x94\\x8c\\x04eval\\x94\\x93\\x94\\x8c*__import__(\'os\').system(\'echo Exploited!\')\\x94\\x85\\x94R\\x94.")\n\n[1]    28027 killed     ipython\n```\nA log file is automatically created, containing everything you need to know:\n```\n$ less /tmp/.secimport/sandbox_pickle.log\n\n  @posix_spawn from /Users/avilumelsky/Downloads/Python-3.10.0/Lib/threading.py\n    DETECTED SHELL:\n        depth=8\n        sandboxed_depth=0\n        sandboxed_module=/Users/avilumelsky/Downloads/Python-3.10.0/Lib/pickle.py  \n\n    TERMINATING SHELL:\n        libsystem_kernel.dylib`__posix_spawn+0xa\n        ...\n                libsystem_kernel.dylib`__posix_spawn+0xa\n                libsystem_c.dylib`system+0x18b\n                python.exe`os_system+0xb3\n    KILLED\n:\n```\nMore examples are available at <a href="docs/EXAMPLES.md">EXAMPLES.md</a>.\n\n<br><br>\n# Roadmap\n- ✔️ Allow/Block list configuration\n- ✔️ Create a .yaml configuration per module in the code\n  - ✔️ Use secimport to compile that yml\n  - ✔️ Create a single dcript policy\n  - ✔️ Run an application with that policy using dtrace, without using `secure_import`\n- ✔️ <b>Add eBPF basic support using bpftrace</b>\n  - ✔️ bpftrace backend tests\n- <b>Extandible Language Template</b>\n  - Implement bpftrace probes for new languages\n- <b>Go support</b> (bpftrace/dtrace hooks)\n  - Implement a template for golang\'s call stack\n- <b>Node support</b> (bpftrace/dtrace hooks)\n  - Implement a template for Node\'s call stack and event loop\n- Multi Process support: Use current_module_str together with thread ID to distinguish between events in different processes\n- Update all linux syscalls in the templates (filesystem, networking, processing) to improve the sandbox blocking of unknowns.\n',
    'author': 'Avi Lumelsky',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/avilum/secimport',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from secimport.backends.common.instrumentation_backend import InstrumentationBackend
from secimport.backends.common.utils import (
    DEFAULT_BACKEND,
    build_module_sandbox_from_yaml_template,
)
from secimport.sandbox_helper import secure_import

imports = secure_import
generate_sandbox_from_yaml = build_module_sandbox_from_yaml_template
__all__ = [
    "imports",
    "secure_import",
    "generate_sandbox_from_yaml",
    "InstrumentationBackend",
]
